--[[
**********************************************************************
MagicDKP Bidder - Addon to handle the DKP Bidding process
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

local MODULE_NAME = "Bidder"
local mod = MagicDKP
local module = mod:NewModule(MODULE_NAME, "AceEvent-3.0", "AceTimer-3.0")
BD = module
local MagicComm = LibStub("MagicComm-1.0")
local GUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)
local tsort = table.sort
local fmt = string.format
local match = string.match
local playerName
local prioSortVal = {}
local defaultOptions = {
   priorities = { "Primary Role", "Secondary Role", "Off Spec" },
   bidTimeout = 60,
   bidMessage = L["Attention: Bidding for [item] now open for [timeout] seconds."],
   bidMessageNoTime = L["Attention: Bidding for [item] now open."],
   bidClosedMessage = L["Bidding for [item] is now closed."],
   bidChannels = {
      CHAT_MSG_WHISPER = true
   },
   allowOverwrite = true,
   allowLowerBid = true,
   autoClose = false,
   hiddenBids = true,
}

local db = {}
local info = {}

local function clear(mm) for id in pairs(mm) do mm[id] = nil end return mm end
local function count(hash) local count = 0 for i in pairs(hash) do count = count + 1 end return count end

function module:OnInitialize()
   -- register defaults and get db storage from the mothership
   db = mod:GetModuleDatabase(MODULE_NAME, defaultOptions, module.options)
   module.bids = {}
   playerName = UnitName("player")
end

function module:StartNewBid(item)
   if not mod:IsLootMaster() then
      mod:error("You're not the loot master. Not starting bid.")
      return
   end
   local itemid = match(item, ".*|Hitem:(%d+):")
   local bidFrame = module.bidFrame   
   if not bidFrame then
      bidFrame = module:CreateBidFrame()
      module.bidFrame = bidFrame
      module.CreateBidFrame = nil -- of to the garbage collector you go
   end
   bidFrame.item:SetText(item, itemid)

   module:BidCompleted(true)   

   if db.bidTimeout > 0 then module.bidTimeout = time() + db.bidTimeout else module.bidTimeout = time() + 10000000 end
   
   module.currentBidItem = item
   clear(module.bids)
   clear(prioSortVal)
--   clear(module.remainingBidders)

   -- Send bid start notice
   info.item = item
   info.timeout = db.bidTimeout
   mod:SendChatMessage(mod:tokenize(db.bidTimeout > 0 and db.bidMessage or db.bidMessageNoTime, info), "RW")
   module:SendBidRequest(item)

   for channel, enabled in pairs(db.bidChannels) do
      if enabled then
	 module:RegisterEvent(channel, "HandleBidMessage")
      end
   end

   if module.bidTimer then module:CancelTimer(module.bidTimer, true) end

   if db.bidTimeout > 0 then
      module.bidTimer = module:ScheduleRepeatingTimer("BidTimerUpdate", 1)
   else
      module:BidTimerUpdate()
   end

   -- Clear old list entries
   for i = 2, #module.scroll.children do
      GUI:Release(module.scroll.children[i])
      module.scroll.children[i] = nil
      module.scroll:DoLayout()
   end

   bidFrame.bidEndButton:SetDisabled(false)

   bidFrame:Show()

   module.lastSorted = "priority"
   module.reverse = false

   for id,prio in pairs(db.priorities) do prioSortVal[prio] = id end
   MagicComm:RegisterListener(module, "MD")
end

local bidRunningRemainingMessage = L["Bid in progress: %d seconds remaining. %d/%d bids."]
local bidRunningMessage = L["Bid in progress. %d/%d bids."]
local bidCompletedMessage = L["Bid completed. %d/%d bids."]

function module:BidTimerUpdate()
   if db.bidTimeout > 0 then
      local remaining = module.bidTimeout - time()
      if remaining <= 0 then
	 module:BidCompleted()
      else
	 module.bidFrame:SetStatusText(bidRunningRemainingMessage:format(remaining, #module.scroll.children-1, GetNumRaidMembers()))
      end
   else
      module.bidFrame:SetStatusText(bidRunningMessage:format(#module.scroll.children-1, GetNumRaidMembers()))
   end
end

function module:BidCompleted(quiet)
   if not module.currentBidItem then return end
   for channel in pairs(db.bidChannels) do
      module:UnregisterEvent(channel)
   end
   module.bidFrame:SetStatusText(bidCompletedMessage:format(#module.scroll.children-1, GetNumRaidMembers()))

   if not quiet then 
      info.item = module.currentBidItem
      mod:SendChatMessage(mod:tokenize(db.bidClosedMessage, info), "GROUP")
   end

   module:CancelTimer(module.bidTimer, true)
   module.bidTimer = nil
   module.bidTimeout = time()
   module.currentBidItem = nil
   MagicComm:UnregisterListener(module, "MD")

   module.bidFrame.bidEndButton:SetDisabled(true)
end

function module:OnDKPResponse(bidder, bid, priority, item, whisper)
   local response = module.bids[bidder]
   if not bid then
      priority = #db.priorities+1
      bid = "---"
   end
   if response then
      if not db.allowOverwrite then return end -- already bid, don't allow it again
      if not db.allowLowerBid and bid < response.bid then return end -- we can't lower bids
      response.widget:SetBid(bid)
   else
      response = { 
	 bidder = bidder,
	 widget = module:GetBidRowWidget(bidder, bid, "")
      }
      module.scroll:AddChild(response.widget)
      module.bids[bidder] = response
   end
   response.message = db.priorities[priority] or whisper or L["Pass"]
   response.whisper = whisper
   response.bid = bid
   response.priority = priority
   response.widget:SetPriority(response.message)
   if db.hiddenBids then
      if not module.bids[playerName] then
	 response.widget:SetBid("(hidden)")
	 response.widget:SetPriority(L["Bid hidden pending own bid"])
      elseif bidder == playerName then
	 for bidder, response in pairs(module.bids) do
	    response.widget:SetPriority(response.message)
	    response.widget:SetBid(response.bid)
	 end
      end
   end
   module:BidDialog_Sort("RESORT");
   if db.autoClose and (#module.scroll.children-1) >= GetNumRaidMembers() then
      module:BidCompleted()
   elseif db.bidTimeout == 0 then
      module:BidTimerUpdate()
   end
end

function module:SendBidRequest(item)
   mod:SetNetworkData("DKPBID", item, db.priorities, db.bidTimeout > 0 and db.bidTimeout)
   mod:SendUrgentMessage("RAID")
end

function module:GetBidRowWidget(player, bid, prio)
   local row = GUI:Create("MDKPRow")
   local _,class = UnitClass(player)
   row:SetData(player, bid, prio, class and RAID_CLASS_COLORS[class])
   return row
end


function findpattern(text, pattern, start)
   local a, b = string.find(text, pattern, start)
   if a and b then 
      return string.sub(text, a, b)
   end
end

function module:HandleBidMessage(channel, msg, sender)
   local bid = tonumber(msg) -- try numeric only bids first
   if not bid then bid = findpattern(msg, "%d+ ") end -- "100 offspec"
   if not bid then bid = findpattern(msg, " %d+") end -- "I bid 100"
   if not bid then return end -- no match, ignoring
   bid = tonumber(bid)

   msg = L["Whisper: "]..msg
   prioSortVal[msg] = #db.priorities+1
   module:OnDKPResponse(sender, bid, #db.priorities+1, module.currentBidItem, msg)
end

module.sortMethods = {
   bid = function(row1, row2) return (tonumber(row1.bid:GetText()) or -1) > (tonumber(row2.bid:GetText()) or -1) end,  
   bidReverse = function(row1, row2) return (tonumber(row1.bid:GetText()) or -1) < (tonumber(row2.bid:GetText()) or -1) end, 
   player = function(row1, row2) return row1.player:GetText() < row2.player:GetText() end, 
   playerReverse = function(row1, row2) return row1.player:GetText() > row2.player:GetText() end, 
   priority = function(row1, row2)
		 local p1, p2 = row1.priority:GetText(), row2.priority:GetText()
		 if p1 == p2 then
		    return module.sortMethods.bid(row1, row2)
		 end
		 return (prioSortVal[p1] or #db.priorities+1) < (prioSortVal[p2] or #db.priorities+1)
	      end, 
   priorityReverse = function(row1, row2)
		 local p1, p2 = row1.priority:GetText(), row2.priority:GetText()
		 if p1 == p2 then
		    return module.sortMethods.bid(row1, row2)
		 end
		 return (prioSortVal[p1] or #db.priorities+1) > (prioSortVal[p2] or #db.priorities+1)
	      end, 
}

function module:Debug() 
   module:OnDKPResponse("Amimage", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Herbalism", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Morasia", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Micke", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Jonas", math.random(1000), math.random(#db.priorities))
end

function module:BidDialog_Sort(event, row)
   local header
   local children = {}
   if event ~= "RESORT" then
      if module.lastSorted == row then
	 module.reverse = not module.reverse
      else
	 module.reverse = nil
      end
      module.lastSorted = row or module.lastSorted
   else
      row = module.lastSorted
   end
   
   if module.reverse then row = row.."Reverse" end
   header = module.scroll.children[1]
   for i = 2, #module.scroll.children do
      children[#children + 1] = module.scroll.children[i]
   end
   table.sort(children, module.sortMethods[row])
   for i = 1, #children do
      module.scroll.children[i+1] = children[i]
   end
   module.scroll:DoLayout()
end

-- Creates the DKP Bid Master UI
function module:CreateBidFrame()
   local function AddSpacer(widget)
      local space = GUI:Create("MDKPSpacer")
      space.width = "fill"
      space.alignoffset = 0
      widget:AddChild(space)
   end

   local frame = GUI:Create("Frame")
   frame:SetTitle(L["DKP Bid Master"])
   frame:SetLayout("Flow")
   frame.width = "fill"
   frame.frame:SetFrameStrata("HIGH")
   frame:SetWidth(440)
   frame:SetHeight(490)
   frame.frame:SetMinResize(440,400)


   local label = GUI:Create("MDKPLabel")
   label:SetText(L["Item:"])
   label:SetWidth(100)
   frame:AddChild(label)
   
   local item = GUI:Create("MDKPItemWidget")
   item:SetWidth(300)
   frame.item = item
   frame:AddChild(item)

   local button = GUI:Create("Button")
   button:SetText(L["End Bidding"])
   button:SetCallback("OnClick", function() module:BidCompleted() end)
   frame.bidEndButton = button
   frame:AddChild(button)

   local scroll = GUI:Create("ScrollFrame")
   scroll:SetLayout("List")
   local header = GUI:Create("MDKPHeader")
   header:SetData("Player", "Bid", "Bid Priority")
   header:SetCallback("OnRowClick", module.BidDialog_Sort)
   
   scroll:AddChild(header)
   module.scroll = scroll
   
   local ig = GUI:Create("InlineGroup")
   ig.width = "fill"
   ig.height = "fill"
   ig:SetLayout("Fill")
   ig:AddChild(scroll)
   frame:AddChild(ig)

   return frame
end


function module:SetProfileParam(var, value, sub)
   local varName = var[#var]
   if sub ~= nil then
      if type(db[varName]) ~= "table" then db[varName] = mod.get() end
      db[varName][value] = sub
   else
      if varName == "priorities" then
	 local newValues = db.varName or {}
	 clear(newValues)
	 for _,prio in ipairs(  { strsplit("\n", value) } ) do 
	    prio = prio:trim()
	    if strlen(prio) > 0 then
	       newValues[#newValues + 1] = prio
	    end
	 end
	 value = newValues
      end
      db[varName] = value
   end
end

function module:GetProfileParam(var, sub)
   local varName = var[#var]
   if sub ~= nil then
      return type(db[varName]) == "table" and db[varName][sub]
   elseif varName == "priorities" then
      return strjoin("\n", unpack(db[varName]))
   else
      return db[varName]
   end
end

function module:OnProfileChanged(newdb)
   db = newdb
end

module.staticMenu = {
   DKPBID = {
      {
	 { 
	    value = "DKPBID",
	    text = L["DKP Bidding"],
	    hasArrow = true,
	    notCheckable = true,
	 }
      }, 
      {
	 {
	    text = L["Clear list and start new bid"],
	    icon = "Interface\\Buttons\\UI-GuildButton-MOTD-Up",
	    func = module.StartNewBid,
	    arg1 = module,
	    notCheckable = true,
	 },
      }
   },
   cancelBid = {
      [2] = {
	 {
	    text = L["Close bidding"],
	    func = module.BidCompleted,
	    arg1 = module,
	    notCheckable = true,
	 }
      }
   }
}

-- config options
module.options = {
   type = "group",
   name = L["DKP Bidder"],
   handler = module,
   set = "SetProfileParam",
   get = "GetProfileParam", 
   args = {
      bidTimeout = {
	 type = "range",
	 name = L["Bid Timeout"],
	 desc = L["Time in seconds before the bid time expires. Set to zero to avoid bid expiration."],
	 min = 0, max = 120, step = 1,
	 order = 20,
	 width="full",
      },
      bidExtension = {
	 type = "range",
	 name = L["Bid Extension"],
	 desc = L["Minimum remaining duration of the bidding after receiving a new bid. This can be used to automatically extend the bid duration after a bid is received. A value of zero disables the feature."],
	 min = 0, max = 60, step = 1,
	 order = 21,
	 hidden = true,
	 width = "full", 
      },
      bidMessage = {
	 type = "input",
	 width="full",
	 name = L["Bid Announce Message"],
	 desc = L["The message sent to the raid or party when a new bid begins. The following token can be used: [item] (the item link) and [timeout] (the bid timeout)."],
	 order = 40,
	 hidden = function() return db.bidTimeout == 0 end
      },
      bidMessageNoTime = {
	 type = "input",
	 width="full",
	 name = L["Bid Announce Message"],
	 desc = L["The message sent to the raid or party when a new bid begins. The following token can be used: [item] (the item link)"],
	 order = 40,
	 hidden = function() return db.bidTimeout > 0 end
      },
      bidClosedMessage = {
	 type = "input",
	 width="full",
	 name = L["Bid Close Message"],
	 desc = L["The message sent to the raid or party when the bidding is closed. The following token can be used: [item]."],
	 order = 40,
      },
      bidChannels = {
	 type = "multiselect",
	 values =  {
	    CHAT_MSG_WHISPER = L["Whisper"],
	    CHAT_MSG_RAID = L["Raid"],
	    CHAT_MSG_SAY = L["Say"], 
	 },
	 name = L["Bid Channels"],
	 desc = L["The channels that DKPBidder will listen to bids on. If no channels are selected, only bids sent by MagicDKP_Client will be recorded."], 
	 order = 50, 
      },
      allowOverwrite = {
	 type = "toggle",
	 name = L["Allow Bid Changes"],
	 desc = L["If enabled whisper bids are allowed to overrwrite previous bids. "], 
	 width = "full",
	 order = 60, 
      },
      allowLowerBid= {
	 type = "toggle",
	 name = L["Allow Lowering Bids"],
	 desc = L["If enabled whisper bids are allowed to lower a previous bid. "],
	 hidden = function() return not db.allowOverwrite end, 
	 width = "full", 
	 order = 60, 
      },
      autoClose = {
	 type = "toggle",
	 name = L["Close Bid Automatically "],
	 desc = L["When enabled, bidding will be closed early if everyone on the bid list has enterd a bid or passed. Passing currently depends on MagicDKP Client - there's no passing mechanism in whisper bidding."],
	 width = "full",
	 order = 70,
      },
      priorities = {
	 type = "input",
	 multiline = true,
	 name = L["Bid Priorities"],
	 desc = L["The list of bid priorities that are sent to the MagicDKP Client. Useful if you separate bids main spec and off spec bids for example. Has no effect on whisper bids."],
	 width = "full", 
      }
   }
}
